@@ -24,6 +24,7 @@ import shlex
 import subprocess
 import pipes
 import imp
+import time
 from StringIO import StringIO
 
 import cmds
@@ -46,7 +47,7 @@ def get_installed_pkgconfigs(config):
             versions = stdout.splitlines()
             if len(versions) == len(pkgs):
                 return dict(zip(pkgs, versions))
-        except OSError:
+        except (subprocess.CalledProcessError, OSError):
             pass
 
         # We have to rather inefficiently repeatedly fork to work around
@@ -59,23 +60,41 @@ def get_installed_pkgconfigs(config):
             stdout = proc.communicate()[0]
             proc.wait()
             pkgversions[pkg] = stdout.strip()
-    except OSError: # pkg-config not installed
+    except (subprocess.CalledProcessError, OSError): # pkg-config not installed
         pass
     return pkgversions
 
-def get_uninstalled_pkgconfigs_and_filenames(uninstalled):
+def get_uninstalled_pkgconfigs(uninstalled):
     uninstalled_pkgconfigs = []
-    uninstalled_filenames = []
-
     for module_name, dep_type, value in uninstalled:
         if dep_type == 'pkgconfig':
             uninstalled_pkgconfigs.append((module_name, value))
-        elif dep_type.lower() == 'path':
+    return uninstalled_pkgconfigs
+
+def get_uninstalled_binaries(uninstalled):
+    uninstalled_binaries = []
+    for module_name, dep_type, value in uninstalled:
+        if dep_type.lower() == 'path':
+            uninstalled_binaries.append((module_name, value))
+    return uninstalled_binaries
+
+def get_uninstalled_c_includes(uninstalled):
+    uninstalled_c_includes = []
+    for module_name, dep_type, value in uninstalled:
+        if dep_type.lower() == 'c_include':
+            uninstalled_c_includes.append((module_name, value))
+    return uninstalled_c_includes
+
+# This function returns uninstalled binaries and C includes as filenames.
+# Backends should use either this OR the functions above.
+def get_uninstalled_filenames(uninstalled):
+    uninstalled_filenames = []
+    for module_name, dep_type, value in uninstalled:
+        if dep_type.lower() == 'path':
             uninstalled_filenames.append((module_name, os.path.join('/usr/bin', value)))
         elif dep_type.lower() == 'c_include':
             uninstalled_filenames.append((module_name, os.path.join('/usr/include', value)))
-
-    return uninstalled_pkgconfigs, uninstalled_filenames
+    return uninstalled_filenames
 
 def systemdependencies_met(module_name, sysdeps, config):
     '''Returns True of the system dependencies are met for module_name'''
@@ -128,11 +147,12 @@ def systemdependencies_met(module_name, sysdeps, config):
         return paths
 
     c_include_search_paths = None
-    for dep_type, value in sysdeps:
+    for dep_type, value, altdeps in sysdeps:
+        dep_met = True
         if dep_type.lower() == 'path':
             if os.path.split(value)[0]:
                 if not os.path.isfile(value) and not os.access(value, os.X_OK):
-                    return False
+                    dep_met = False
             else:
                 pathdirs = set(os.environ.get('PATH', '').split(os.pathsep))
                 pathdirs.update(['/sbin', '/usr/sbin'])
@@ -141,7 +161,7 @@ def systemdependencies_met(module_name, sysdeps, config):
                     if os.path.isfile(filename) and os.access(filename, os.X_OK):
                         break
                 else:
-                    return False
+                    dep_met = False
         elif dep_type.lower() == 'c_include':
             if c_include_search_paths is None:
                 c_include_search_paths = get_c_include_search_paths(config)
@@ -152,28 +172,39 @@ def systemdependencies_met(module_name, sysdeps, config):
                     found = True
                     break
             if not found:
-                return False
+                dep_met = False
 
         elif dep_type == 'python2':
             try:
                 imp.find_module(value)
             except:
-                return False
+                dep_met = False
 
         elif dep_type == 'xml':
-            for d in os.environ['XDG_DATA_DIRS'].split(':'):
-                xml_catalog = os.path.join(d, 'xml', 'catalog')
-                if os.path.exists(xml_catalog):
-                    break
-            else:
-                xml_catalog = '/etc/xml/catalog'
+            xml_catalog = '/etc/xml/catalog'
+
+            if not os.path.exists(xml_catalog):
+                for d in os.environ['XDG_DATA_DIRS'].split(':'):
+                    xml_catalog = os.path.join(d, 'xml', 'catalog')
+                    if os.path.exists(xml_catalog):
+                        break
 
             try:
                 # no xmlcatalog installed will (correctly) fail the check
                 subprocess.check_output(['xmlcatalog', xml_catalog, value])
 
             except:
-                return False
+                dep_met = False
+
+        # check alternative dependencies
+        if not dep_met and altdeps:
+            for altdep in altdeps:
+                if systemdependencies_met(module_name, [ altdep ], config):
+                    dep_met = True
+                    break
+
+        if not dep_met:
+            return False
 
     return True
 
@@ -184,9 +215,9 @@ class SystemInstall(object):
         elif cmds.has_command('sudo'):
             self._root_command_prefix_args = ['sudo']
         else:
-            raise SystemExit, _('No suitable root privilege command found; you should install "pkexec"')
+            raise SystemExit(_('No suitable root privilege command found; you should install "sudo" or "pkexec" (or the system package that provides it)'))
 
-    def install(self, uninstalled):
+    def install(self, uninstalled, assume_yes):
         """Takes a list of pkg-config identifiers and uses a system-specific method to install them."""
         raise NotImplementedError()
 
@@ -205,7 +236,6 @@ PK_FILTER_ENUM_NEWEST = 1 << 16
 PK_FILTER_ENUM_ARCH = 1 << 18
 PK_TRANSACTION_FLAG_ENUM_ONLY_TRUSTED = 1 << 1
 
-# NOTE: This class is unfinished
 class PKSystemInstall(SystemInstall):
     def __init__(self):
         SystemInstall.__init__(self)
@@ -221,72 +251,57 @@ class PKSystemInstall(SystemInstall):
 
     def _get_new_transaction(self):
         if self._loop is None:
-            import glib
+            try:
+                import glib
+            except:
+                try:
+                    from gi.repository import GLib as glib
+                except:
+                    raise SystemExit(_('Error: python-gobject package not found.'))
             self._loop = glib.MainLoop()
         if self._sysbus is None:
-            import dbus.glib
+            try:
+                import dbus.glib
+            except:
+                raise SystemExit(_('Error: dbus-python package not found.'))
             import dbus
             self._dbus = dbus
             self._sysbus = dbus.SystemBus()
         if self._pkdbus is None:
             self._pkdbus = dbus.Interface(self._sysbus.get_object('org.freedesktop.PackageKit',
-                                              '/org/freedesktop/PackageKit'),
-                            'org.freedesktop.PackageKit')
+                                                                  '/org/freedesktop/PackageKit'),
+                                          'org.freedesktop.PackageKit')
             properties = dbus.Interface(self._pkdbus, 'org.freedesktop.DBus.Properties')
-            self._pk_major = properties.Get('org.freedesktop.PackageKit', 'VersionMajor')
-            self._pk_minor = properties.Get('org.freedesktop.PackageKit', 'VersionMinor')
-        if self._pk_major == 1 or (self._pk_major == 0 and self._pk_minor >= 8):
-            txn_path = self._pkdbus.CreateTransaction()
-            txn = self._sysbus.get_object('org.freedesktop.PackageKit', txn_path)
-        else:
-            tid = self._pkdbus.GetTid()
-            txn = self._sysbus.get_object('org.freedesktop.PackageKit', tid)
+        txn_path = self._pkdbus.CreateTransaction()
+        txn = self._sysbus.get_object('org.freedesktop.PackageKit', txn_path)
         txn_tx = self._dbus.Interface(txn, 'org.freedesktop.PackageKit.Transaction')
         txn.connect_to_signal('Message', self._on_pk_message)
         txn.connect_to_signal('ErrorCode', self._on_pk_error)
         txn.connect_to_signal('Destroy', lambda *args: self._loop.quit())
         return txn_tx, txn
 
-    def install(self, uninstalled):
-        uninstalled_pkgconfigs, uninstalled_filenames = get_uninstalled_pkgconfigs_and_filenames(uninstalled)
+    def install(self, uninstalled, assume_yes):
+        logging.info(_('Computing packages to install. This might be slow. Please wait.'))
         pk_package_ids = set()
-
+        uninstalled_pkgconfigs = get_uninstalled_pkgconfigs(uninstalled)
         if uninstalled_pkgconfigs:
             txn_tx, txn = self._get_new_transaction()
             txn.connect_to_signal('Package', lambda info, pkid, summary: pk_package_ids.add(pkid))
-            if self._pk_major == 1 or (self._pk_major == 0 and self._pk_minor >= 9):
-                # PackageKit 1.0.x or 0.9.x
-                txn_tx.WhatProvides(PK_FILTER_ENUM_ARCH | PK_FILTER_ENUM_NEWEST |
-                                    PK_FILTER_ENUM_NOT_INSTALLED,
-                                    ['pkgconfig(%s)' % pkg for modname, pkg in
-                                     uninstalled_pkgconfigs])
-            elif self._pk_major == 0 and self._pk_minor == 8:
-                # PackageKit 0.8.x
-                txn_tx.WhatProvides(PK_FILTER_ENUM_ARCH | PK_FILTER_ENUM_NEWEST |
-                                    PK_FILTER_ENUM_NOT_INSTALLED,
-                                    PK_PROVIDES_ANY,
-                                    ['pkgconfig(%s)' % pkg for modname, pkg in
-                                     uninstalled_pkgconfigs])
-            else:
-                # PackageKit 0.7.x and older
-                txn_tx.WhatProvides('arch;newest;~installed', 'any',
-                                    ['pkgconfig(%s)' % pkg for modname, pkg in
-                                     uninstalled_pkgconfigs])
+            txn_tx.WhatProvides(PK_FILTER_ENUM_ARCH | PK_FILTER_ENUM_NEWEST |
+                                PK_FILTER_ENUM_NOT_INSTALLED,
+                                ['pkgconfig(%s)' % pkg for modname, pkg in
+                                 uninstalled_pkgconfigs])
             self._loop.run()
             del txn, txn_tx
 
+        uninstalled_filenames = get_uninstalled_filenames(uninstalled)
         if uninstalled_filenames:
             txn_tx, txn = self._get_new_transaction()
             txn.connect_to_signal('Package', lambda info, pkid, summary: pk_package_ids.add(pkid))
-            if self._pk_major == 1 or (self._pk_major == 0 and self._pk_minor >= 8):
-                txn_tx.SearchFiles(PK_FILTER_ENUM_ARCH | PK_FILTER_ENUM_NEWEST |
-                                   PK_FILTER_ENUM_NOT_INSTALLED,
-                                   [pkg for modname, pkg in
-                                    uninstalled_filenames])
-            else:
-                txn_tx.SearchFiles('arch;newest;~installed',
-                                   [pkg for modname, pkg in
-                                    uninstalled_filenames])
+            txn_tx.SearchFiles(PK_FILTER_ENUM_ARCH | PK_FILTER_ENUM_NEWEST |
+                               PK_FILTER_ENUM_NOT_INSTALLED,
+                               [pkg for modname, pkg in
+                                uninstalled_filenames])
             self._loop.run()
             del txn, txn_tx
 
@@ -300,19 +315,10 @@ class PKSystemInstall(SystemInstall):
             return
 
         logging.info(_('Installing:\n  %s' % ('\n  '.join(pk_package_ids, ))))
+        logging.info(_('This might take a very long time. Do not turn off your computer. You can run `pkmon\' to monitor progress.'))
 
         txn_tx, txn = self._get_new_transaction()
-        if self._pk_major == 1 or (self._pk_major == 0 and self._pk_minor >= 8):
-            # Using OnlyTrusted might break package installation on rawhide,
-            # where packages are unsigned, but this prevents users of normal
-            # distros with signed packages from seeing security warnings. It
-            # would be better to simulate the transaction first to decide
-            # whether OnlyTrusted will work before using it. See
-            # http://www.freedesktop.org/software/PackageKit/gtk-doc/introduction-ideas-transactions.html
-            txn_tx.InstallPackages(PK_TRANSACTION_FLAG_ENUM_ONLY_TRUSTED, pk_package_ids)
-        else:
-            # PackageKit 0.7.x and older
-            txn_tx.InstallPackages(True, pk_package_ids)
+        txn_tx.InstallPackages(PK_TRANSACTION_FLAG_ENUM_ONLY_TRUSTED, pk_package_ids)
         self._loop.run()
 
         logging.info(_('Complete!'))
@@ -321,39 +327,82 @@ class PKSystemInstall(SystemInstall):
     def detect(cls):
         return cmds.has_command('pkcon')
 
-class YumSystemInstall(SystemInstall):
+class PacmanSystemInstall(SystemInstall):
     def __init__(self):
         SystemInstall.__init__(self)
+        self._pacman_install_args = ['pacman', '-S', '--asdeps', '--needed', '--quiet', '--noconfirm']
 
-    def install(self, uninstalled):
-        uninstalled_pkgconfigs, uninstalled_filenames = get_uninstalled_pkgconfigs_and_filenames(uninstalled)
-        logging.info(_('Using yum to install packages.  Please wait.'))
+    def _maybe_update_pkgfile(self):
+        if not cmds.has_command('pkgfile'):
+            logging.info(_('pkgfile not found, automatically installing'))
+            if subprocess.call(self._root_command_prefix_args + self._pacman_install_args + ['pkgfile',]):
+                logging.error(_('Failed to install pkgfile'))
+                raise SystemExit
 
-        if len(uninstalled_pkgconfigs) + len(uninstalled_filenames) > 0:
-            logging.info(_('Installing:\n  %(pkgs)s') %
-                         {'pkgs': '\n  '.join([modname for modname, pkg in
-                                               uninstalled_pkgconfigs +
-                                               uninstalled_filenames])})
-            args = self._root_command_prefix_args + ['yum', '-y', 'install']
-            args.extend(['pkgconfig(%s)' % pkg for modname, pkg in
-                         uninstalled_pkgconfigs])
-            args.extend([pkg for modname, pkg in uninstalled_filenames])
-            subprocess.check_call(args)
-        else:
+        # Update the pkgfile cache if it is older than 1 day.
+        cacheexists = bool(os.listdir('/var/cache/pkgfile'))
+        if not cacheexists or os.stat('/var/cache/pkgfile').st_mtime < time.time() - 86400:
+            logging.info(_('pkgfile cache is old or doesn\'t exist, automatically updating'))
+            result = subprocess.call(self._root_command_prefix_args + ['pkgfile', '--update'])
+            if result and not cacheexists:
+                logging.error(_('Failed to create pkgfile cache'))
+                raise SystemExit
+            elif result:
+                logging.warning(_('Failed to update pkgfile cache'))
+            else:
+                logging.info(_('Successfully updated pkgfile cache'))
+
+    def install(self, uninstalled, assume_yes):
+        uninstalled_pkgconfigs = get_uninstalled_pkgconfigs(uninstalled)
+        uninstalled_filenames = get_uninstalled_filenames(uninstalled)
+        logging.info(_('Using pacman to install packages.  Please wait.'))
+        package_names = set()
+
+        if not uninstalled_filenames and not uninstalled_pkgconfigs:
             logging.info(_('Nothing to install'))
+            return
+
+        self._maybe_update_pkgfile()
+
+        for name, pkgconfig in uninstalled_pkgconfigs:
+            # Just throw the pkgconfigs in the normal file list
+            uninstalled_filenames.append((None, '/usr/lib/pkgconfig/%s.pc' %pkgconfig))
+
+        for name, filename in uninstalled_filenames:
+            try:
+                result = subprocess.check_output(['pkgfile', '--raw', filename])
+                if result:
+                    package_names.add(result.split('\n')[0])
+            except subprocess.CalledProcessError:
+                logging.warning(_('Provider for "%s" was not found, ignoring' %(name if name else filename)))
+
+        if not package_names:
+            logging.info(_('Nothing to install'))
+            return
+
+        logging.info('Installing:\n  %s' %('\n  '.join(package_names)))
+        if subprocess.call(self._root_command_prefix_args + self._pacman_install_args + list(package_names)):
+            logging.error(_('Install failed'))
+        else:
+            logging.info(_('Completed!'))
 
     @classmethod
     def detect(cls):
-        return cmds.has_command('yum')
-
+        if cmds.has_command('pacman'):
+            return True
+        return False
 
 class AptSystemInstall(SystemInstall):
     def __init__(self):
         SystemInstall.__init__(self)
 
-    def _get_package_for(self, filename):
-        proc = subprocess.Popen(['apt-file', 'search', filename],
-                                stdout=subprocess.PIPE, close_fds=True)
+    def _get_package_for(self, filename, exact_match):
+        if exact_match:
+            proc = subprocess.Popen(['apt-file', '--fixed-string', 'search', filename],
+                                    stdout=subprocess.PIPE, close_fds=True)
+        else:
+            proc = subprocess.Popen(['apt-file', 'search', filename],
+                                    stdout=subprocess.PIPE, close_fds=True)
         stdout = proc.communicate()[0]
         if proc.returncode != 0:
             return None
@@ -363,33 +412,71 @@ class AptSystemInstall(SystemInstall):
                 continue
             name = parts[0]
             path = parts[1]
-            # No idea why the LSB has forks of the pkg-config files
-            if path.find('/lsb3') != -1:
+            # Ignore copies of the pkg-config files that are not from the
+            # libraries.
+            if '/lsb3' in path:
                 continue
-            
+            if '/emscripten' in path:
+                continue
+
             # otherwise for now, just take the first match
             return name
 
-    def install(self, uninstalled):
-        uninstalled_pkgconfigs, uninstalled_filenames = get_uninstalled_pkgconfigs_and_filenames(uninstalled)
-        logging.info(_('Using apt-file to search for providers; this may be slow.  Please wait.'))
+    def _try_append_native_package(self, modname, filename, native_packages, exact_match):
+        native_pkg = self._get_package_for(filename, exact_match)
+        if native_pkg:
+            native_packages.append(native_pkg)
+            return True
+        return False
+
+    def _append_native_package_or_warn(self, modname, filename, native_packages, exact_match):
+        if not self._try_append_native_package(modname, filename, native_packages, exact_match):
+            logging.info(_('No native package found for %(id)s '
+                           '(%(filename)s)') % {'id'       : modname,
+                                                'filename' : filename})
+
+    def _install_packages(self, native_packages, assume_yes):
+        logging.info(_('Installing: %(pkgs)s') % {'pkgs': ' '.join(native_packages)})
+        apt_cmd_line = ['apt-get']
+        if assume_yes is True:
+            apt_cmd_line += ['--assume-yes']
+        apt_cmd_line += ['install']
+        args = self._root_command_prefix_args + apt_cmd_line
+        args.extend(native_packages)
+        subprocess.check_call(args)
+
+    def install(self, uninstalled, assume_yes):
+        logging.info(_('Using apt-file to search for providers; this may be extremely slow. Please wait. Patience!'))
         native_packages = []
+
         pkgconfigs = [(modname, '/%s.pc' % pkg) for modname, pkg in
-                      uninstalled_pkgconfigs]
-        for modname, filename in pkgconfigs + uninstalled_filenames:
-            native_pkg = self._get_package_for(filename)
-            if native_pkg:
-                native_packages.append(native_pkg)
-            else:
-                logging.info(_('No native package found for %(id)s '
-                               '(%(filename)s)') % {'id'       : modname,
-                                                   'filename' : filename})
+                      get_uninstalled_pkgconfigs(uninstalled)]
+        for modname, filename in pkgconfigs:
+            self._append_native_package_or_warn(modname, filename, native_packages, False)
+
+        binaries = [(modname, '/usr/bin/%s' % pkg) for modname, pkg in
+                    get_uninstalled_binaries(uninstalled)]
+        for modname, filename in binaries:
+            self._append_native_package_or_warn(modname, filename, native_packages, True)
+
+        # Get multiarch include directory, e.g. /usr/include/x86_64-linux-gnu
+        multiarch = None
+        try:
+            multiarch = subprocess.check_output(['gcc', '-print-multiarch']).strip()
+        except:
+            # Really need GCC to continue. Yes, this is fragile.
+            self._install_packages(['gcc'])
+            multiarch = subprocess.check_output(['gcc', '-print-multiarch']).strip()
+
+        c_includes = get_uninstalled_c_includes(uninstalled)
+        for modname, filename in c_includes:
+            # Try multiarch first, so we print the non-multiarch location on failure.
+            if (multiarch == None or
+                not self._try_append_native_package(modname, '/usr/include/%s/%s' % (multiarch, filename), native_packages, True)):
+                self._append_native_package_or_warn(modname, '/usr/include/%s' % filename, native_packages, True)
 
         if native_packages:
-            logging.info(_('Installing: %(pkgs)s') % {'pkgs': ' '.join(native_packages)})
-            args = self._root_command_prefix_args + ['apt-get', 'install']
-            args.extend(native_packages)
-            subprocess.check_call(args)
+            self._install_packages(native_packages, assume_yes=assume_yes)
         else:
             logging.info(_('Nothing to install'))
 
@@ -398,7 +485,7 @@ class AptSystemInstall(SystemInstall):
         return cmds.has_command('apt-file')
 
 # Ordered from best to worst
-_classes = [AptSystemInstall, PKSystemInstall, YumSystemInstall]
+_classes = [AptSystemInstall, PacmanSystemInstall, PKSystemInstall]
 
 if __name__ == '__main__':
     logging.basicConfig(level=logging.INFO)
